
from bson import ObjectId
from fastapi import HTTPException
from typing import List

from repositories.user_repo import (
    create_user,
    get_user,
    get_users,
    update_user,
    delete_user,
)
from schemas.user_schema import UserCreate, UserUpdate, UserOut,UserBase,UserRefresh
from security.hash import check_password
from security.encrypting_jwt import create_jwt_member_token
from repositories.tokens_repo import add_refresh_tokens, add_access_tokens, accessTokenCreate,accessTokenOut,refreshTokenCreate
from repositories.tokens_repo import get_refresh_tokens,get_access_tokens,delete_access_token,delete_refresh_token,delete_all_tokens_with_user_id

async def add_user(user_data: UserCreate) -> UserOut:
    """adds an entry of UserCreate to the database and returns an object

    Returns:
        _type_: UserOut
    """
    user =  await get_user(filter_dict={"email":user_data.email})
    if user==None:
        new_user= await create_user(user_data)
        access_token = await add_access_tokens(token_data=accessTokenCreate(userId=new_user.id))
        refresh_token  = await add_refresh_tokens(token_data=refreshTokenCreate(userId=new_user.id,previousAccessToken=access_token.accesstoken))
        new_user.password=""
        new_user.access_token= access_token.accesstoken 
        new_user.refresh_token = refresh_token.refreshtoken
        return new_user
    else:
        raise HTTPException(status_code=409,detail="User Already exists")

async def authenticate_user(user_data:UserBase )->UserOut:
    user = await get_user(filter_dict={"email":user_data.email})

    if user != None:
        if check_password(password=user_data.password,hashed=user.password ):
            user.password=""
            access_token = await add_access_tokens(token_data=accessTokenCreate(userId=user.id))
            refresh_token  = await add_refresh_tokens(token_data=refreshTokenCreate(userId=user.id,previousAccessToken=access_token.accesstoken))
            user.access_token= access_token.accesstoken 
            user.refresh_token = refresh_token.refreshtoken
            return user
        else:
            raise HTTPException(status_code=401, detail="Unathorized, Invalid Login credentials")
    else:
        raise HTTPException(status_code=404,detail="User not found")

async def refresh_user_tokens_reduce_number_of_logins(user_refresh_data:UserRefresh,expired_access_token):
    refreshObj= await get_refresh_tokens(user_refresh_data.refresh_token)
    if refreshObj:
        if refreshObj.previousAccessToken==expired_access_token:
            user = await get_user(filter_dict={"_id":ObjectId(refreshObj.userId)})
            
            if user!= None:
                    access_token = await add_access_tokens(token_data=accessTokenCreate(userId=user.id))
                    refresh_token  = await add_refresh_tokens(token_data=refreshTokenCreate(userId=user.id,previousAccessToken=access_token.accesstoken))
                    user.access_token= access_token.accesstoken 
                    user.refresh_token = refresh_token.refreshtoken
                    await delete_access_token(accessToken=expired_access_token)
                    await delete_refresh_token(refreshToken=user_refresh_data.refresh_token)
                    return user
     
        await delete_refresh_token(refreshToken=user_refresh_data.refresh_token)
        await delete_access_token(accessToken=expired_access_token)
  
    raise HTTPException(status_code=404,detail="Invalid refresh token ")  
        
async def remove_user(user_id: str):
    """deletes a field from the database and removes UserCreateobject 

    Raises:
        HTTPException 400: Invalid user ID format
        HTTPException 404:  User not found
    """
    if not ObjectId.is_valid(user_id):
        raise HTTPException(status_code=400, detail="Invalid user ID format")

    filter_dict = {"_id": ObjectId(user_id)}
    result = await delete_user(filter_dict)
    await delete_all_tokens_with_user_id(userId=user_id)

    if result.deleted_count == 0:
        raise HTTPException(status_code=404, detail="User not found")


async def retrieve_user_by_user_id(id: str) -> UserOut:
    """Retrieves user object based specific Id 

    Raises:
        HTTPException 404(not found): if  User not found in the db
        HTTPException 400(bad request): if  Invalid user ID format

    Returns:
        _type_: UserOut
    """
    if not ObjectId.is_valid(id):
        raise HTTPException(status_code=400, detail="Invalid user ID format")

    filter_dict = {"_id": ObjectId(id)}
    result = await get_user(filter_dict)

    if not result:
        raise HTTPException(status_code=404, detail="User not found")

    return result


async def retrieve_users(start=0,stop=100) -> List[UserOut]:
    """Retrieves UserOut Objects in a list

    Returns:
        _type_: UserOut
    """
    return await get_users(start=start,stop=stop)


async def update_user_by_id(user_id: str, user_data: UserUpdate) -> UserOut:
    """_summary_

    Raises:
        HTTPException 404(not found): if User not found or update failed
        HTTPException 400(not found): Invalid user ID format

    Returns:
        _type_: UserOut
    """
    if not ObjectId.is_valid(user_id):
        raise HTTPException(status_code=400, detail="Invalid user ID format")

    filter_dict = {"_id": ObjectId(user_id)}
    result = await update_user(filter_dict, user_data)

    if not result:
        raise HTTPException(status_code=404, detail="User not found or update failed")

    return result




from schemas.imports import *
from pydantic import Field
import time
from security.hash import hash_password
class UserBase(BaseModel):
    # Add other fields here 
    email:EmailStr
    password:str | bytes
    pass

class UserRefresh(BaseModel):
    # Add other fields here 
    refresh_token:str
    pass


class UserCreate(UserBase):
    # Add other fields here 
    date_created: int = Field(default_factory=lambda: int(time.time()))
    last_updated: int = Field(default_factory=lambda: int(time.time()))
    @model_validator(mode='after')
    def obscure_password(self):
        self.password=hash_password(self.password)
        return self
class UserUpdate(BaseModel):
    # Add other fields here 
    last_updated: int = Field(default_factory=lambda: int(time.time()))

class UserOut(UserBase):
    # Add other fields here 
    id: Optional[str] =None
    date_created: Optional[int] = None
    last_updated: Optional[int] = None
    refresh_token: Optional[str] =None
    access_token:Optional[str]=None
    @model_validator(mode='before')
    def set_dynamic_values(cls,values):
        if isinstance(values,dict):
            values['id']= str(values.get('_id'))
            return values
      
            
    class Config:
        from_attributes = True
        populate_by_name = True
        arbitrary_types_allowed = True
        json_encoders = {
            ObjectId: str
        }


import redis
import os

cache_db = redis.Redis(
    host=os.getenv("REDIS_HOST"),
    port=int(os.getenv("REDIS_PORT")),
    username=os.getenv("REDIS_USERNAME"),
    password=os.getenv("REDIS_PASSWORD"),
   decode_responses=True
)


import click
from fasterapi.scaffolder.generate_project import create_project
from fasterapi.scaffolder.generate_crud import create_crud_file
from fasterapi.scaffolder.generate_schema import create_schema_file
from fasterapi.scaffolder.generate_service import create_service_file
from fasterapi.scaffolder.generate_route import create_route_file,get_highest_numbered_api_version,get_latest_modified_api_version
from fasterapi.__version__ import __version__
from fasterapi.scaffolder.mount_routes import update_main_routes
from fasterapi.scaffolder.generate_tokens_repo import create_token_file
from fasterapi.scaffolder.generate_dependencies import create_dependency_file
import subprocess

@click.group()
@click.version_option(__version__, '--version', '-v', message='FasterAPI version %(version)s')
def cli():
    """⚡ FasterAPI CLI — Scaffold FastAPI apps with ease"""
    pass


@cli.command()
@click.argument("name")
def new(name):
    """Create a new FastAPI project."""
    create_project(name)

@cli.command()
@click.argument("name")
def make_crud(name):
    """
    Generate CRUD repository functions for a schema.

    Requires that a schema with the given NAME already exists.

    \b
    ✅ Good usage:
        fasterapi make-crud user
        fasterapi make-crud product
        fasterapi make-crud order

    ❌ Bad usage:
        fasterapi make-crud        # Missing name
        fasterapi make-crud User   # Avoid capital letters
        fasterapi make-crud user profile  # Too many arguments

    Notes:
        - The corresponding schema must already exist before running this.
    """
    create_crud_file(name)
    
@cli.command()
@click.argument("name")
def make_schema(name):
    """
    Generate Pydantic class templates for a schema.

    \b
    ✅ Good usage:
        fasterapi make-schema user
        fasterapi make-schema product
        fasterapi make-schema order

    ❌ Bad usage:
        fasterapi make-schema        # Missing name
        fasterapi make-schema User   # Avoid capital letters
        fasterapi make-schema user profile  # Too many arguments

    Notes:
        - The schema will be created in the appropriate project folder.
    """
    create_schema_file(name)
    

@cli.command()
@click.argument("name")
def make_service(name):
    """
    Generate Python service templates to interact with a schema and repository.

    \b
    ✅ Good usage:
        fasterapi make-service user
        fasterapi make-service product
        fasterapi make-service order

    ❌ Bad usage:
        fasterapi make-service        # Missing name
        fasterapi make-service User   # Avoid capital letters
        fasterapi make-service user profile  # Too many arguments

    Notes:
      
        - A matching schema and repository should already exist.
    """
    create_service_file(name)
    
@cli.command()
def mount():
    """
    Mount all API routes into the main FastAPI application file.

    This command scans the `api/v*` directories and automatically updates
    your `main.py` (the entrypoint file) to include the discovered routes.

    \b
    ✅ Good usage:
        fasterapi mount
        
    ❌ Bad usage:
        fasterapi mount user       # This command takes no arguments
        fasterapi mount --help me  # Use only 'fasterapi mount' or 'fasterapi mount --help'
        
    Notes:
        - Run this after generating new routes (e.g., with `make-route`).
        - Existing imports and route mounts will be preserved.
    """
    try:
        update_main_routes()
        click.secho("✅ Routes successfully mounted into main.py.", fg="green")
    except Exception as e:
        click.secho(f"❌ Failed to mount routes: {e}", fg="red")
        raise click.Abort()
    
@cli.command(name="run-d")
def run_d():
    """
    Run the FastAPI development server with Uvicorn (auto-reload enabled).

    \b
    ✅ Good usage:
        fasterapi run-d

    ❌ Bad usage:
        fasterapi run-d extra       # This command takes no arguments

    Notes:
        - This is equivalent to running:
            uvicorn main:app --reload
        - Ensure you have Uvicorn installed (pip install uvicorn).
    """
    try:
        subprocess.run(["uvicorn", "main:app", "--reload"], check=True)
    except FileNotFoundError:
        click.secho("❌ Uvicorn is not installed. Run 'pip install uvicorn' and try again.", fg="red")
    except subprocess.CalledProcessError as e:
        click.secho(f"❌ Failed to start server: {e}", fg="red")
        raise click.Abort()


@cli.command()
def update():
    """
    Upgrade the FasterAPI CLI to the latest version.

    \b
    ✅ Good usage:
        fasterapi update

    ❌ Bad usage:
        fasterapi update extra   # This command takes no arguments

    Notes:
        - This will run:
            pip install --upgrade nats-fasterapi
        - Ensure you have pip available in your PATH.
    """
    try:
        subprocess.run(
            ["pip", "install", "--upgrade", "nats-fasterapi"],
            check=True
        )
        click.secho("✅ FasterAPI has been upgraded to the latest version!", fg="green")
    except FileNotFoundError:
        click.secho("❌ Pip is not installed or not found in PATH.", fg="red", err=True)
        raise click.Abort()
    except subprocess.CalledProcessError:
        click.secho("❌ Failed to upgrade FasterAPI. Please try again.", fg="red", err=True)
        raise click.Abort()

@cli.command()
@click.argument("name")
@click.option(
    "--version-mode",
    type=click.Choice(["latest-modified", "highest-number"], case_sensitive=True),
    required=False,
    help="Choose API versioning strategy: 'latest-modified' or 'highest-number'.",
)
@click.option(
    "-y", "--yes",
    is_flag=True,
    help="Skip prompts and use default options (non-interactive mode).",
)
def make_route(name, version_mode, yes):
    """
    Generate an API route file for a given schema.

    \b
    ✅ Good usage:
        fasterapi make-route user --version-mode latest-modified
        fasterapi make-route product --version-mode highest-number
        fasterapi make-route order                # Will ask interactively
        fasterapi make-route order -y             # Skips prompt, uses default

    ❌ Bad usage:
        fasterapi make-route                      # Missing name
        fasterapi make-route user foo             # Invalid version-mode

    Notes:
        - NAME should be a lowercase schema name (e.g., 'user').
        - If no --version-mode is provided:
            → Prompt interactively (unless -y is used).
            → Defaults to 'highest-number' when skipped.
    """
    if not version_mode:
        if yes:
            version_mode = "highest-number"
            click.secho("⚠️ No version mode provided. Using default: highest-number", fg="yellow")
        else:
            version_mode = click.prompt(
                "Select version mode",
                type=click.Choice(["latest-modified", "highest-number"]),
                default="highest-number",
                show_choices=True,
            )

    if version_mode == "latest-modified":
        version = get_latest_modified_api_version()
    else:  # "highest-number"
        version = get_highest_numbered_api_version()

    click.secho(f"📌 Selected API version: {version}", fg="cyan")
    create_route_file(name, version)
    click.secho(f"✅ Route for '{name}' created successfully.", fg="green")

 
@cli.command()
@click.argument("roles", nargs=-1)
def make_token_repo(roles):
    """
    Generate a token repository based on roles.

    \b
    ✅ Good usage:
        fasterapi make-token-repo admin user
        fasterapi make-token-repo staff guest-editor
        fasterapi make-token-repo                 # Defaults will be used

    ❌ Bad usage:
        fasterapi make-token-repo admin 1 3 , @! user staff extra-role arg  

    Notes:
        - Provide one or more roles (space-separated).
        - If no roles are provided, defaults will be used:
          admin, user, staff, guest-editor
    """
    if not roles:
        roles = ["admin", "user"]
        click.secho(
            f"⚠️ No roles provided. Using default roles: {', '.join(roles)}",
            fg="yellow"
        )

    create_token_file(roles)
    click.secho(
        f"✅ Token repository generated successfully for roles: {', '.join(roles)}",
        fg="green"
    )


@cli.command()
def make_token_deps():
    """
    Generate the token dependency file from the token repository.

    \b
    ✅ Good usage:
        fasterapi make-token-deps

    ❌ Bad usage:
        fasterapi make-token-deps extra-arg   # This command takes no arguments

    Notes:
        - Run this after creating a token repository (make-token-repo).
        - The generated file contains dependency utilities for authentication.
    """
    try:
        create_dependency_file()
        click.secho("✅ Token dependency file generated successfully.", fg="green")
    except Exception as e:
        click.secho(f"❌ Failed to generate token dependencies: {e}", fg="red")
        raise click.Abort()

@cli.command(name="git-push-auto")
def git_push_auto():
    """
    Automates a three-step Git workflow: add, commit, and push.

    \b
    ✅ Good usage:
        fasterapi git-push-auto

    ❌ Bad usage:
        fasterapi git-push-auto extra    # This command takes no arguments

    Notes:
        - This is equivalent to running:
            git add .
            git commit -m "automated commit"
            git push origin master
    """
    try:
        click.secho("Adding all changes...", fg="cyan")
        subprocess.run(["git", "add", "."], check=True)

        click.secho("Committing with message 'automated commit'...", fg="cyan")
        subprocess.run(["git", "commit", "-m", "automated commit"], check=True)

        click.secho("Pushing to origin master...", fg="cyan")
        subprocess.run(["git", "push", "origin", "master"], check=True)

        click.secho("✅ Git workflow completed successfully!", fg="green")

    except FileNotFoundError:
        click.secho("❌ Git is not installed or not in your PATH. Please install Git and try again.", fg="red")
        raise click.Abort()

    except subprocess.CalledProcessError as e:
        click.secho(f"❌ Failed to complete Git workflow: {e}", fg="red")
        click.secho("A Git command failed. Please check your repository status and try the commands manually.", fg="red")
        raise click.Abort()



if __name__ == "__main__":
    cli()




from fastapi import APIRouter, HTTPException, Query, status, Path,Depends
from typing import List
from schemas.response_schema import APIResponse
from schemas.tokens_schema import accessTokenOut
from schemas.user_schema import (
    UserCreate,
    UserOut,
    UserBase,
    UserUpdate,
    UserRefresh,
)
from services.user_service import (
    add_user,
    remove_user,
    retrieve_users,
    authenticate_user,
    retrieve_user_by_user_id,
    update_user,
    refresh_user_tokens_reduce_number_of_logins,

)
from security.auth import verify_token,verify_token_to_refresh
router = APIRouter(prefix="/users", tags=["Users"])

@router.get("/{start}/{stop}", response_model=APIResponse[List[UserOut]],response_model_exclude_none=True,dependencies=[Depends(verify_token)])
async def list_users(start:int= 0, stop:int=100):
    items = await retrieve_users(start=0,stop=100)
    return APIResponse(status_code=200, data=items, detail="Fetched successfully")

@router.get("/me", response_model=APIResponse[UserOut],dependencies=[Depends(verify_token)],response_model_exclude_none=True)
async def get_my_users(token:accessTokenOut = Depends(verify_token)):
    items = await retrieve_user_by_user_id(id=token.userId)
    return APIResponse(status_code=200, data=items, detail="users items fetched")



@router.post("/signup", response_model=APIResponse[UserOut])
async def signup_new_user(user_data:UserBase):
    new_user = UserCreate(**user_data.model_dump())
    items = await add_user(user_data=new_user)
    return APIResponse(status_code=200, data=items, detail="Fetched successfully")


@router.post("/login", response_model=APIResponse[UserOut])
async def login_user(user_data:UserBase):
    items = await authenticate_user(user_data=user_data)
    return APIResponse(status_code=200, data=items, detail="Fetched successfully")


@router.post("/refesh",response_model=APIResponse[UserOut],dependencies=[Depends(verify_token_to_refresh)])
async def refresh_user_tokens(user_data:UserRefresh,token:accessTokenOut = Depends(verify_token_to_refresh)):
    
    items= await refresh_user_tokens_reduce_number_of_logins(user_refresh_data=user_data,expired_access_token=token.accesstoken)

    return APIResponse(status_code=200, data=items, detail="users items fetched")


@router.delete("/account",dependencies=[Depends(verify_token)])
async def delete_user_account(token:accessTokenOut = Depends(verify_token)):
    result = await remove_user(user_id=token.userId)
    return result



from pymongo import ReturnDocument
from core.database import db
from fastapi import HTTPException,status
from typing import List,Optional
from schemas.user_schema import UserUpdate, UserCreate, UserOut

async def create_user(user_data: UserCreate) -> UserOut:
    user_dict = user_data.model_dump()
    result =await db.users.insert_one(user_dict)
    result = await db.users.find_one(filter={"_id":result.inserted_id})
  
    returnable_result = UserOut(**result)
    return returnable_result

async def get_user(filter_dict: dict) -> Optional[UserOut]:
    try:
        result = await db.users.find_one(filter_dict)

        if result is None:
            return None

        return UserOut(**result)

    except Exception as e:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=f"An error occurred while fetching user: {str(e)}"
        )
    
async def get_users(filter_dict: dict = {},start=0,stop=100) -> List[UserOut]:
    try:
        if filter_dict is None:
            filter_dict = {}

        cursor = (db.users.find(filter_dict)
        .skip(start)
        .limit(stop - start)
        )
        user_list = []

        async for doc in cursor:
            userObj =UserOut(**doc)
            userObj.password=None
            user_list.append(userObj)
        
        return user_list

    except Exception as e:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=f"An error occurred while fetching users: {str(e)}"
        )
async def update_user(filter_dict: dict, user_data: UserUpdate) -> UserOut:
    result = await db.users.find_one_and_update(
        filter_dict,
        {"$set": user_data.model_dump()},
        return_document=ReturnDocument.AFTER
    )
    returnable_result = UserOut(**result)
    return returnable_result

async def delete_user(filter_dict: dict):
    return await db.users.delete_one(filter_dict)


from fastapi import FastAPI, Request, HTTPException
from fastapi.responses import JSONResponse
from fastapi.middleware.cors import CORSMiddleware
from starlette.middleware.base import BaseHTTPMiddleware
from starlette.responses import Response
from limits import Limiter, RateLimitItemPerMinute
from limits.storage import MemoryStorage
from limits.errors import RateLimitExceeded
from schemas.response_schema import APIResponse
import time   

# TODO: Modify this if you want to use the time it took for a request to be completed in anyway like maybe calculating the average speed of each individual endpoint
class RequestTimingMiddleware(BaseHTTPMiddleware):
    async def dispatch(self, request, call_next):
        # Record the start time before processing the request
        start_time = time.time()
        
        # Process the request and get the response
        response = await call_next(request)
        
        # Calculate the time taken to process the request
        process_time = time.time() - start_time
        
        # You can log the time or set it in the response headers
        response.headers['X-Process-Time'] = str(process_time)
        
        # Optionally, print it for logging purposes
        print(f"Request to {request.url} took {process_time:.6f} seconds")
        
        return response
    
# Create the FastAPI app
app = FastAPI()
app.add_middleware(RequestTimingMiddleware)
# Setup limiter
storage = MemoryStorage()
limiter = Limiter(storage)

# TODO: Define rate limits per user type Change the logic here to be more dynamic or not
RATE_LIMITS = {
    "free": RateLimitItemPerMinute(5),
    "premium": RateLimitItemPerMinute(20),
    "admin": RateLimitItemPerMinute(100),
}

# TODO: Dummy user resolution function (replace with function to actually get user from the request object)
def get_user_type(request: Request) -> tuple[str, str]:
    """
    Return a tuple of (user_identifier, user_type)
    You can extract from JWT, headers, or session.
    """
    user_id = request.headers.get("X-User-ID", "anonymous")
    user_type = request.headers.get("X-User-Type", "free").lower()
    return user_id, user_type if user_type in RATE_LIMITS else "free"

class RateLimitingMiddleware(BaseHTTPMiddleware):
    async def dispatch(self, request: Request, call_next):
        user_id, user_type = get_user_type(request)
        rate_limit_rule = RATE_LIMITS[user_type]

        try:
            # Use user_id as the rate limit key
            limiter.hit(rate_limit_rule, user_id)

            response = await call_next(request)
            return response

        except RateLimitExceeded:
            # Get time until the limit resets
            _, reset_time = limiter.get_window_stats(rate_limit_rule, user_id)
            current_time = int(time.time())
            seconds_until_reset = max(reset_time - current_time, 0)

            return JSONResponse(
                status_code=429,
                content=APIResponse(
                    status_code=429,
                    data={
                        "retry_after_seconds": seconds_until_reset,
                        "user_type": user_type
                    },
                    detail="Too Many Requests"
                ).dict()
            )
            

# Add the middleware to the app
# ||||||||||||||||||||||||||||||||||||

# app.add_middleware(RateLimitingMiddleware)

# ||||||||||||||||||||||||||||||||||||

# Add CORS middleware (be cautious in production)
app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],  # Adjust for production
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

# Custom exception handler for HTTPExceptions
@app.exception_handler(HTTPException)
async def custom_http_exception_handler(request: Request, exc: HTTPException):
    return JSONResponse(
        status_code=exc.status_code,
        content=APIResponse(
            status_code=exc.status_code,
            data=None,
            detail=exc.detail,
        ).dict()
    )

# Simple test route
@app.get("/")
def read_root():
    return {"message": "Hello from FasterAPI!"}

# Health check route
@app.get("/health")
async def health_check():
    return {"status": "healthy"}
